#include "MyScalarBarActor"
#include "vtkScalarBarActorInternal.h"

#include "vtkCellArray.h"
#include "vtkCellData.h"
#include "vtkColor.h"
#include "vtkCoordinate.h"
#include "vtkFloatArray.h"
#include "vtkPointData.h"
#include "vtkImageData.h"
#include "vtkMath.h"
#include "vtkObjectFactory.h"
#include "vtkPolyData.h"
#include "vtkPolyDataMapper2D.h"
#include "vtkProperty2D.h"
#include "vtkRenderer.h"
#include "vtkRenderWindow.h"
#include "vtkScalarsToColors.h"
#include "vtkSmartPointer.h"
#include "vtkTextActor.h"
#include "vtkTextProperty.h"
#include "vtkTexture.h"
#include "vtkTexturedActor2D.h"
#include "vtkViewport.h"
#include "vtkWindow.h"
#include "vtkMathTextUtilities.h"

#include <vector>
#include <set>
#include <map>

#include <stdio.h> // for snprintf

vtkStandardNewMacro(MyScalarBarActor)

MyScalarBarActor::MyScalarBarActor() : vtkScalarBarActor()
{

}

//----------------------------------------------------------------------------
MyScalarBarActor::~MyScalarBarActor()
{

}

//----------------------------------------------------------------------------
void MyScalarBarActor::RebuildLayout(vtkViewport* viewport)
{
  vtkScalarBarActor::RebuildLayout(viewport);
}

//----------------------------------------------------------------------------
int MyScalarBarActor::RenderOverlay(vtkViewport* viewport)
{
    int renderedSomething = 0;

    // Is the viewport's RenderWindow capturing GL2PS-special props? We'll need
    // to handle this specially to get the texture to show up right.
    if (vtkRenderer *renderer = vtkRenderer::SafeDownCast(viewport))
    {
        if (vtkRenderWindow *renderWindow = renderer->GetRenderWindow())
        {
            if (renderWindow->GetCapturingGL2PSSpecialProps())
            {
                renderer->CaptureGL2PSSpecialProp(this);
            }
        }
    }

    // Everything is built, just have to render
    if (this->DrawBackground)
    {
        renderedSomething += this->BackgroundActor->RenderOverlay(viewport);
    }

    if (this->UseOpacity && this->DrawColorBar)
    {
        renderedSomething += this->TextureActor->RenderOverlay(viewport);
    }

    // Draw either the scalar bar (non-indexed mode) or
    // the annotated value boxes (indexed mode).
    if (!this->LookupTable->GetIndexedLookup())
    {
        if (this->DrawColorBar)
        {
            renderedSomething += this->ScalarBarActor->RenderOverlay(viewport);
        }

        if (this->DrawTickLabels)
        {
            vtkScalarBarActorInternal::ActorVector::iterator it;
            for (
                 it = this->P->TextActors.begin();
                 it != this->P->TextActors.end();
                 ++it)
            {
                renderedSomething += (*it)->RenderOverlay(viewport);
            }
        }
    }
    else if (this->DrawColorBar)
    {
        renderedSomething +=
                this->P->AnnotationBoxesActor->RenderOverlay(viewport);
    }

    if (this->DrawNanAnnotation)
    {
        renderedSomething +=
                this->P->NanSwatchActor->RenderOverlay(viewport);
    }

    if (this->DrawFrame)
    {
        renderedSomething += this->FrameActor->RenderOverlay(viewport);
    }

    if (this->Title != NULL)
    {
        renderedSomething += this->TitleActor->RenderOverlay(viewport);
    }

    if (this->DrawAnnotations)
    {
        if (!this->P->AnnotationLabels.empty())
        {
            if(this->P->AnnotationLeadersActor != NULL)
            {
                renderedSomething +=
                        this->P->AnnotationLeadersActor->RenderOverlay( viewport );

                for (vtkScalarBarActorInternal::ActorVector::size_type i = 0; i < this->P->AnnotationLabels.size(); ++ i )
                {
                    renderedSomething +=
                            this->P->AnnotationLabels[i]->RenderOverlay( viewport );
                }
            }
        }
    }

    renderedSomething = (renderedSomething > 0)?(1):(0);
    return renderedSomething;
}

void MyScalarBarActor::DrawBoxes()
{
    vtkScalarBarActor::DrawBoxes();
}

//----------------------------------------------------------------------------
void MyScalarBarActor::ComputeFrame()
{
    vtkScalarBarActor::ComputeFrame();
}

//----------------------------------------------------------------------------
void MyScalarBarActor::ComputeScalarBarThickness()
{
    vtkScalarBarActor::ComputeScalarBarThickness();
}

//----------------------------------------------------------------------------
void MyScalarBarActor::LayoutNanSwatch()
{
    vtkScalarBarActor::LayoutNanSwatch();
}

//----------------------------------------------------------------------------
void MyScalarBarActor::PrepareTitleText()
{
    vtkScalarBarActor::PrepareTitleText();
}

//----------------------------------------------------------------------------
void MyScalarBarActor::LayoutTitle()
{

    vtkScalarBarActor::LayoutTitle();
}

//-----------------------------------------------------------------------------
void MyScalarBarActor::ComputeScalarBarLength()
{
    vtkScalarBarActor::ComputeScalarBarLength();
}

//-----------------------------------------------------------------------------
void MyScalarBarActor::LayoutTicks()
{
    vtkScalarBarActor::LayoutTicks();
}

//----------------------------------------------------------------------------
void MyScalarBarActor::LayoutAnnotations()
{
    vtkScalarBarActor::LayoutAnnotations();
}

//----------------------------------------------------------------------------
void MyScalarBarActor::ConfigureFrame()
{
    vtkScalarBarActor::ConfigureFrame();
}

//----------------------------------------------------------------------------
void MyScalarBarActor::ConfigureScalarBar()
{
    std::cout << "Configure Scalar Bar" << std::endl;
  vtkScalarsToColors* lut = this->LookupTable;
  const double* range = lut->GetRange();
  this->P->NumColors = lut->GetIndexedLookup() ?
    lut->GetNumberOfAnnotatedValues() : this->MaximumNumberOfColors;
  //this->P->NumColors += 2;
  this->P->NumSwatches = this->P->NumColors + (this->DrawNanAnnotation ? 1 : 0);
  int numPts = 2 * (this->P->NumColors + 1) + (this->DrawNanAnnotation ? 4 : 0);
  this->P->SwatchPts = vtkPoints::New();
  this->P->SwatchPts->SetNumberOfPoints(numPts+4);
  this->P->Polys = vtkCellArray::New();
  this->P->Polys->Allocate(
    this->P->Polys->EstimateSize(this->P->NumSwatches,4));
  this->P->SwatchColors = vtkUnsignedCharArray::New();

  unsigned int nComponents = ((this->UseOpacity) ? 4 : 3);
  this->P->SwatchColors->SetNumberOfComponents(nComponents);
  this->P->SwatchColors->SetNumberOfTuples(this->P->NumSwatches+2);

  this->ScalarBarActor->SetProperty(this->GetProperty());
  this->ScalarBar->Initialize();
  this->ScalarBar->SetPoints(this->P->SwatchPts);
  this->ScalarBar->SetPolys(this->P->Polys);
  this->ScalarBar->GetCellData()->SetScalars(this->P->SwatchColors);
  this->P->SwatchPts->Delete();
  this->P->Polys->Delete();
  this->P->SwatchColors->Delete();

  std::cout << "\t\t NumColors: " << this->P->NumColors << std::endl;
  std::cout << "\t\t NumSwatches: " << this->P->NumSwatches << std::endl;
  std::cout << "\t\t SwatchPts: " << this->P->SwatchPts->GetNumberOfPoints() << std::endl;
  std::cout << "\t\t Polys: " << this->P->Polys->GetNumberOfCells() << std::endl;
  std::cout << "\t\t SwatchColors: " << this->P->SwatchColors->GetNumberOfTuples() << std::endl;
  std::cout << "\t\t NumPts: " << numPts << std::endl;

  double delta =
    static_cast<double>(this->P->ScalarBarBox.Size[1]) / (this->P->NumColors+2);
  double x[3];
  x[2] = 0.;
  x[this->P->TL[0]] = this->P->ScalarBarBox.Posn[this->P->TL[0]];
  x[this->P->TL[1]] = this->P->ScalarBarBox.Posn[this->P->TL[1]] - 10 * delta;
  this->P->SwatchPts->SetPoint(0, x);
  std::cout << "\t\t\t INDEX: " << 0 << " ---- " << x[0] << ", " << x[1] << ", " << x[2] << std::endl;


  x[this->P->TL[0]] = this->P->ScalarBarBox.Posn[this->P->TL[0]] + this->P->ScalarBarBox.Size[0];
  //x[this->P->TL[1]] = this->P->ScalarBarBox.Posn[this->P->TL[1]];
  this->P->SwatchPts->SetPoint(1, x);
  std::cout << "\t\t\t INDEX: " << 1 << " ---- " << x[0] << ", " << x[1] << ", " << x[2] << std::endl;

  int pointId = 0;
  for ( ; pointId < numPts / 2 - (this->DrawNanAnnotation ? 2 : 0); ++pointId)
    {
    x[this->P->TL[0]] = this->P->ScalarBarBox.Posn[this->P->TL[0]];
    x[this->P->TL[1]] = this->P->ScalarBarBox.Posn[this->P->TL[1]] + pointId * delta;
    this->P->SwatchPts->SetPoint(2 * (pointId+1), x);
    //std::cout << "\t\t\t INDEX: " << 2 * (pointId+1) << " ---- " << x[0] << ", " << x[1] << ", " << x[2] << std::endl;

    x[this->P->TL[0]] = this->P->ScalarBarBox.Posn[this->P->TL[0]] +
      this->P->ScalarBarBox.Size[0];
    //std::cout << "\t\t\t INDEX: " << 2 * (pointId+1) + 1 << " ---- " << x[0] << ", " << x[1] << ", " << x[2] << std::endl;
    this->P->SwatchPts->SetPoint(2 * (pointId+1) + 1, x);
    }
  x[this->P->TL[0]] = this->P->ScalarBarBox.Posn[this->P->TL[0]];
  x[this->P->TL[1]] = this->P->ScalarBarBox.Posn[this->P->TL[1]] + (pointId+10) * delta;
  this->P->SwatchPts->SetPoint(2 * (pointId+1), x);
  std::cout << "\t\t\t INDEX: " << 2 * (pointId+1) << " ---- " << x[0] << ", " << x[1] << ", " << x[2] << std::endl;

  x[this->P->TL[0]] = this->P->ScalarBarBox.Posn[this->P->TL[0]] + this->P->ScalarBarBox.Size[0];
  this->P->SwatchPts->SetPoint(2 * (pointId+1) + 1, x);
  std::cout << "\t\t\t INDEX: " << 2 * (pointId+1) + 1 << " ---- " << x[0] << ", " << x[1] << ", " << x[2] << std::endl;
//  x[this->P->TL[1]] = this->P->ScalarBarBox.Posn[this->P->TL[1]];

  //polygons & cell colors
  unsigned char* rgb;
  double rgba[4];
  vtkIdType ptIds[4];

  for (int i = 0; i < (this->P->NumColors+2); ++i)
    {
    ptIds[0] = (2 * i);
    ptIds[1] = ptIds[0] + 1;
    ptIds[2] = ptIds[1] + 2;
    ptIds[3] = ptIds[0] + 2;

    this->P->Polys->InsertNextCell(4,ptIds);
    double rgbval;
    if (this->LookupTable->UsingLogScale())
      {
      rgbval = log10(range[0]) +
        i * (log10(range[1]) - log10(range[0])) / this->P->NumColors;
      rgbval = pow(10.0, rgbval);
      }
    else
      {
      rgbval = range[0] + (range[1] - range[0]) *
        (i / static_cast<double>(this->P->NumColors));
      }
    lut->GetColor(rgbval, rgba);
    rgba[3] = lut->GetOpacity(rgbval);
    //write into array directly
    rgb = this->P->SwatchColors->GetPointer(nComponents * i);

    if(i == 0)
    {
        rgba[0] = 0;    rgba[1] = 0;    rgba[2] = 0;
    }
    else if(i == (this->P->NumColors - 1))
    {
        rgba[0] = 1;    rgba[1] = 1;    rgba[2] = 1;
    }
    rgb[0] = static_cast<unsigned char>(rgba[0] * 255.);
    rgb[1] = static_cast<unsigned char>(rgba[1] * 255.);
    rgb[2] = static_cast<unsigned char>(rgba[2] * 255.);
    if (this->P->SwatchColors->GetNumberOfComponents() > 3)
      {
      rgb[3] = static_cast<unsigned char>(this->UseOpacity ? rgba[3] * 255. : 255.);
      }

//    std::cout << "ID: " << i << "rgbval: " << rgbval << ", RGB: " << static_cast<int>(rgb[0]) << ", " <<
//                                             static_cast<int>(rgb[1]) << ", " <<
//                                             static_cast<int>(rgb[2]) << std::endl;
    }

  std::cout << "Final number of polys: " << this->P->Polys->GetNumberOfCells() << std::endl;

  // Set up a texture actor as an alternative to the 1-quad-per-color
  // scalar bar polydata.
  vtkPoints* texturePoints = vtkPoints::New();
  texturePoints->SetNumberOfPoints(4);
  this->TexturePolyData->SetPoints(texturePoints);
  texturePoints->SetPoint(0, 0.0, 0.0, 0.0);

  double p1[2], p2[2];
  p1[0] = this->P->ScalarBarBox.Posn[0];
  p1[1] = this->P->ScalarBarBox.Posn[1];
  p2[0] = p1[0] + this->P->ScalarBarBox.Size[this->P->TL[0]];
  p2[1] = p1[1] + this->P->ScalarBarBox.Size[this->P->TL[1]];

  texturePoints->SetPoint(0, p1[0], p1[1], 0.0);
  texturePoints->SetPoint(1, p2[0], p1[1], 0.0);
  texturePoints->SetPoint(2, p2[0], p2[1], 0.0);
  texturePoints->SetPoint(3, p1[0], p2[1], 0.0);
  texturePoints->Delete();

  std::cout << "P0: " << p1[0] << ", " << p1[1] << std::endl;
  std::cout << "P1: " << p2[0] << ", " << p2[1] << std::endl;

  double barWidth = this->P->ScalarBarBox.Size[this->P->TL[0]];
  double barHeight = this->P->ScalarBarBox.Size[this->P->TL[1]];

  std::cout << "\t\t Bar width: " << barWidth << ", " << barHeight << std::endl;
  vtkDataArray* tc = this->TexturePolyData->GetPointData()->GetTCoords();
  tc->SetTuple2(1, barWidth / this->TextureGridWidth, 0.0);
  tc->SetTuple2(2, barWidth / this->TextureGridWidth,
                barHeight / this->TextureGridWidth);
  tc->SetTuple2(3, 0.0, barHeight / this->TextureGridWidth);
}

//----------------------------------------------------------------------------
void MyScalarBarActor::ConfigureTitle()
{
    vtkScalarBarActor::ConfigureTitle();
}

//-----------------------------------------------------------------------------
void MyScalarBarActor::ConfigureTicks()
{
    vtkScalarBarActor::ConfigureTicks();
}

//-----------------------------------------------------------------------------
void MyScalarBarActor::ConfigureNanSwatch()
{
    vtkScalarBarActor::ConfigureNanSwatch();
}

//----------------------------------------------------------------------------
void MyScalarBarActor::ConfigureAnnotations()
{
    vtkScalarBarActor::ConfigureAnnotations();
}

//----------------------------------------------------------------------------
void MyScalarBarActor::FreeLayoutStorage()
{
  vtkScalarBarActor::FreeLayoutStorage();
}

//----------------------------------------------------------------------------
void MyScalarBarActor::SizeTitle(double* titleSize,
                                  int* size,
                                  vtkViewport* viewport)
{
  vtkScalarBarActor::SizeTitle(titleSize, size, viewport);
}
